from dataclasses import dataclass
from typing import overload, Union, Tuple
from math import pi
from build123d import (
    Vector,
    VectorLike,
    RotationLike,
    Location,
    Plane,
    Edge,
    Face,
    Wire,
    Shape,
    CenterOf,
)


class Mate:
    @overload
    def __init__(
        self,
        origin: VectorLike = Vector(0, 0, 0),
        x_dir: VectorLike = Vector(1, 0, 0),
        z_dir: VectorLike = Vector(0, 0, 1),
    ):
        ...

    @overload
    def __init__(self, shape: Union[Face, Wire, Edge]):
        ...

    def __init__(self, *args):

        if len(args) == 1 and isinstance(args[0], Mate):
            val = args[0]

            self.origin = val.origin
            self.x_dir = val.x_dir
            self.y_dir = val.y_dir
            self.z_dir = val.z_dir

        elif len(args) == 1 and isinstance(args[0], Shape):
            val = args[0]

            self.origin = val.center(CenterOf.BOUNDING_BOX)

            if isinstance(val, (Edge, Wire)):
                self.z_dir = val.normal()

                vertices = val.vertices()
                if len(vertices) == 1:  # e.g. a single closed spline
                    # Use the vector defined by the vertex and the origin as x direction
                    self.x_dir = vertices[0] - self.origin
                else:
                    # Use the vector defined by the first two vertices as x direction
                    self.x_dir = vertices[1] - vertices[0]

            elif isinstance(val, Face):
                self.z_dir = val.normal_at(val.center())

                # x_dir will be derived from the local coord system of the underlying plane
                xd = val._geom_adaptor().Position().XDirection()
                self.x_dir = Vector(xd.X(), xd.Y(), xd.Z())

            else:
                raise ValueError("Needs a Face, Wire, or an Edge")

        elif len(args) <= 3 and all([isinstance(a, (Vector, tuple)) for a in args]):
            self.origin = Vector(args[0])
            self.x_dir = Vector(args[1]).normalized()
            self.z_dir = Vector(args[2]).normalized()

        else:
            raise ValueError(f"Needs a Face, Wire, or an Edge, not {args}")

        self.y_dir = self.z_dir.cross(self.x_dir)

    def __repr__(self) -> str:
        c = lambda v: f"({v.X:.2f}, {v.Y:.2f}, {v.Z:.2f})"
        return f"Mate(origin={c(self.origin)}, x_dir={c(self.x_dir)}, z_dir={c(self.z_dir)})"

    @property
    def loc(self) -> Location:
        return Location(self.to_plane())

    @classmethod
    def from_plane(cls, plane: Plane) -> "Mate":
        return cls(plane.origin, plane.x_dir, plane.z_dir)

    def to_plane(self) -> Plane:
        return Plane(self.origin, self.x_dir, self.z_dir)

    def translate(self, vector: VectorLike):
        """
        Translate with a given direction scaled by dist
        :param axis: the direction to translate
        """
        self.origin = self.origin + Vector(vector)

    def translated(self, vector: VectorLike) -> "Mate":
        """
        Return a mate, translated with a given direction scaled by dist
        :param axis: the direction to translate
        """
        mate = Mate(self)  # copy mate
        mate.translate(vector)
        return mate

    def rotate(self, angles: RotationLike):
        """
        Rotate with a given angles in the order x, y, z
        :param angles: (x, y, z) angles to rotate (in degrees)
        """
        plane = self.to_plane().rotated(angles)
        self.origin = plane.origin
        self.x_dir = plane.x_dir
        self.z_dir = plane.z_dir
        self.y_dir = self.z_dir.cross(self.x_dir)

    def rotated(self, angles: RotationLike) -> "Mate":
        """
        Return a new mate, rotated with a given angles in the order x, y, z
        :param angles: (x, y, z) angles to rotate (in degrees)
        :return: self
        """
        mate = Mate(self)  # copy mate
        mate.rotate(angles)
        return mate

    def move(self, loc: Location):
        """
        Move by the given Location
        :param loc: The Location object to move the mate
        """

        def move(origin: Vector, vec: Vector, loc: Location) -> Tuple[Vector, Vector]:
            reloc = Edge.make_line(origin, origin + vec).moved(loc)
            v1, v2 = reloc.start_point(), reloc.end_point()
            return v1, v2 - v1

        origin, x_dir = move(self.origin, self.x_dir, loc)
        _, z_dir = move(self.origin, self.z_dir, loc)
        return Mate(origin, x_dir, z_dir)

    def moved(self, loc: Location) -> "Mate":
        """
        Return a new mate moved by the given Location
        :param loc: The Location object to move the mate
        :return: Mate
        """
        mate = Mate(self)  # copy mate
        mate.move(loc)
        return mate

    # Some short cuts

    # def rx(self, angle: float) -> "Mate":
    #     """
    #     Rotate with a given angle around x axis
    #     :param angle: angle to rotate (in degrees)
    #     :return: self
    #     """
    #     self.rotate((angle / 180 * pi, 0, 0))

    # def ry(self, angle: float) -> "Mate":
    #     """
    #     Rotate with a given angle around y axis
    #     :param angle: angle to rotate (in degrees)
    #     :return: self
    #     """
    #     self.rotate((0, angle / 180 * pi, 0))

    # def rz(self, angle: float) -> "Mate":
    #     """
    #     Rotate with a given angle around z axis
    #     :param angle: angle to rotate (in degrees)
    #     :return: self
    #     """
    #     self.rotate((0, 0, angle / 180 * pi))

    # def tx(self, dist: float) -> "Mate":
    #     """
    #     Translate with a given distance along x axis
    #     :param dist: distance to translate
    #     :return: self
    #     """
    #     self.translate(self.x_dir * dist)

    # def ty(self, dist: float) -> "Mate":
    #     """
    #     Translate with a given distance along y axis
    #     :param dist: distance to translate
    #     :return: self
    #     """
    #     self.translate(self.y_dir * dist)
    #     return self

    # def tz(self, dist: float) -> "Mate":
    #     """
    #     Translate with a given distance along z axis
    #     :param dist: distance to translate
    #     :return: self
    #     """
    #     self.translate(self.z_dir * dist)
    #     return self
